@@ -1,15 +1,16 @@
 # See COPYING file in this directory.
 # Some code originally from django-boundaryservice
-
-from urllib import unquote
+from __future__ import unicode_literals
+try:
+    from urllib.parse import unquote
+except ImportError:
+    from urllib import unquote
 
 from django.contrib.gis.db.models import GeometryField
-try:
-    import json as simplejson
-except ImportError: # < Python 2.6
-    from django.utils import simplejson
 from django.contrib.gis.geos import GEOSGeometry
 
+import json
+
 from tastypie.fields import ApiField, CharField
 from tastypie import resources
 
@@ -26,7 +27,7 @@ class GeometryApiField(ApiField):
         value = super(GeometryApiField, self).hydrate(bundle)
         if value is None:
             return value
-        return simplejson.dumps(value)
+        return json.dumps(value)
 
     def dehydrate(self, obj, for_list=False):
         return self.convert(super(GeometryApiField, self).dehydrate(obj))
@@ -41,7 +42,7 @@ class GeometryApiField(ApiField):
         # Get ready-made geojson serialization and then convert it _back_ to
         # a Python object so that tastypie can serialize it as part of the
         # bundle.
-        return simplejson.loads(value.geojson)
+        return json.loads(value.geojson)
 
 
 class ModelResource(resources.ModelResource):
